<?php if ( ! defined('BASEPATH')) exit('No direct script access allowed');

/**
 * Base class for fields on a page.
 */
class Field {
  public $name;        // string- field name
  public $type;        // string- one of the below types
  public $default;     // string
  public $required;    // boolean
  public $address;     // string- where values are located in JSON
  public $hidemaster;  // int- if >0, show a more/less button
  public $hidegroup;   // int- if >0, hide if button on corresponding hidemaster header is clicked
  public $json;        // array- all the data for the page
  public $numlines;    // int- number of lines high to make the field in the grid
  public $entryList;   // string - entry list - one of the following types (none, start, cont, end)

/**
 * Factory method. Takes "type" from json and returns a field of the appropriate subclass.
 * @param string $name - field name
 * @param array $json  - portion of config JSON for this field
 * @return Field
 */
  static function createFromJson($name,$json) {
    switch ($json["type"])  {
      case "header":
        $field=new Field_Header($json,$name);
        break;
      case "boolean":
        $field=new Field_Boolean($json,$name);
        break;
      case "int":
        $field=new Field_Int($json,$name);
        break;
      case "string":
        $field=new Field_String($json,$name);
        break;
      case "password":
        $field=new Field_Password($json,$name);
        break;
      case "ipaddress":
        $field=new Field_Ipaddress($json,$name);
        break;
      case "enum":
        $field=new Field_Enum($json,$name);
        break;
      case "multi":
        $field=new Field_Multi($json,$name);
        break;
      default:
        $field=new Field($json,$name);
        break;
    }
    return $field;
  }

/**
 * Common values to all fields.
 * @param array $json
 */
  function __construct($json,$name) {
    $this->name=$name;
    $this->json=$json;
    $this->type=$json["type"];
    $this->default=array_key_exists("default",$json)?$json["default"]:"";
    $this->required=array_key_exists("required",$json)?$json["required"]:0;
    $this->address=array_key_exists("address",$json)?$json["address"]:"";
    $this->hidemaster=array_key_exists("hidemaster",$json)?$json["hidemaster"]:0;
    $this->hidegroup=array_key_exists("hidegroup",$json)?$json["hidegroup"]:0;
    $this->numlines=array_key_exists("numlines",$json)?$json["numlines"]:1;
    // HBB - added 12/27/2012
    $this->entryList=array_key_exists("entryList",$json)?$json["entryList"]:"none";
  }

/**
 * Dump row of schema
 */
  function schema_table()
  {
    $type=get_class($this);
    $nicename = $this->display_name();
    echo "<td>$nicename</td><td>$this->type</td><td>$type</td><td>default:$this->default</td>";
  }

/**
 * Gets the translated name of the field. Some subclasses override this to provide formatting.
 * @return string
 */
  function display_name() {

    return lang_d($this->name);
  }


/**
 * Takes a value and formats it based on the subclass's attributes.
 * @param type $value
 * @return string
 */
  function format_value($value) {
    return $value;
  }

/**
 * Checks value against constraints, returns error message or NULL if good.
 * @param int $portnum
 * @return string
 */
  function validate($portnum=NULL) {
    return NULL;
  }

/**
 * Gets value for this field from POST
 * @param int $portnum
 * @return value - NULL if value is not in POST
 */
  function get_value_from_post($portnum=NULL)
  {
    if (array_key_exists($this->name,$_POST)) {
      $value=$_POST[$this->name];
      if ($portnum!==NULL) {
        if (is_array($value) && array_key_exists($portnum,$value)) {
          $value=$value[$portnum];
        } else {
          $value=NULL;
        }
      }
    } else {
      $value=NULL;
    }
    return $value;
  }

/**
 * Gets the HTML name to use to match get_value_from_post
 * @param int $portnum
 * @return string
 */
  function get_html_name($portnum=NULL) {
    $name=$this->name;
    if ($portnum!==NULL) {
      $name.="[$portnum]";
    }
    return $name;
  }

/**
 * Creates the HTML to edit this field.
 * @param array $json - pass the json array so we can get the value
 * @param int $portnum
 * @return string
 */
  function edit_html(&$json,$portnum=NULL)
  {
    $value=htmlspecialchars($this->read_value($json,$portnum),ENT_QUOTES);
    $name=$this->get_html_name($portnum);
    return "<input type='text' name='$name' value='$value'>\n";
  }

/**
 * Pulls a value out of the JSON array based on the address.
 * @param array $json     - Passed by reference to save memory
 * @param int $portnumber - optional port number, replaces [*] in address.
 * @return value
 */
  function read_value(&$json,$portnumber=0,$sessionnumber=0) {
    if ($this->address) {
      $a=explode('.',$this->address);
      $j=&$json;
      foreach ($a AS $index) {
        if ($index=='[*]') $index=$portnumber;
        if ($index=='[?]') $index=$sessionnumber;
        if (array_key_exists($index, $j)) {
          $j=&$j[$index];
        } else {
          return $this->default;
        }
      }
      return $j;
    } else return NULL;
  }

/**
 * Writes a value into the JSON array based on the address.
 *
 * @param array $json     - Passed by reference because we're going to change it
 * @param type $value     - value to save
 * @param int $portnumber - optional port number, replaces [*] in address.
 * @return boolean        - true if value written
 */
  function write_value(&$json,$value,$portnumber=0) {
    if ($this->address) {
      $a=explode('.',$this->address);
      $j=&$json;
      $num=count($a);
      for ($i=0;$i<$num; $i++) {
        $index=$a[$i];
        if ($index=='[*]') $index=$portnumber;
        if (array_key_exists($index, $j)) {
          $j=&$j[$index]; // follow by reference so we can write into the JSON array
        } else {
          if ($i<$num-1) {
            $j[$index]=array(); // create next level in json
            $j=&$j[$index];
          } else {
            $j[$index]=$value; // create leaf in json
            return true;
          }
        }
      }
      $j=$value;  // At this point, $j is a reference to the value in the $json array where the value is stored
    }
    return true;
  }
}


/* Field_Header ****************************************************************/

class Field_Header extends Field {
  public $level;  // dept
  function __construct($json,$name) {
    parent::__construct($json,$name);
    $this->level = array_key_exists("level",$json)?$json["level"]:1;
  }

/**
 * overrides to display level instead of default
 */
  function schema_table() {
    $l=$this->level;
    $type=get_class($this);
    $nicename = $this->display_name();
    echo "<td>$nicename</td><td>$this->type</td><td>$type</td><td>$this->level</td>";
  }

/**
 * Overrides to make sure it returns an empty value
 * @param type $value
 * @return string
 */
  function format_value($value) {
    $l=$this->level;
    return "<h$l>&nbsp;</h$l>";
  }

/**
 * Adds <h> tag of appropriate level
 * @return string
 */
  function display_name() {
    $l=$this->level;
    return "<h$l>".parent::display_name()."</h$l>";
  }

/**
 * No edit html for no value
 * @return string
 */
  function edit_html() {
    $l=$this->level;
    return "<h$l>&nbsp;</h$l>";
  }

}

/* Field_Boolean ***************************************************************/

class Field_Boolean extends Field {
  function __construct($json,$name) {
    parent::__construct($json,$name);
  }

/**
 * Returns string "true" or "false" instead of whatever comes out of JSON
 * @param boolean $value
 * @return string
 */
  function format_value($value) {
    return $value ? 'true' : 'false';
  }

/**
 * Makes a checkbox
 * @param boolean $value
 * @param int $portnum
 * @return string
 */
  function edit_html(&$json,$portnum=NULL) {
    $value=$this->read_value($json,$portnum);
    $name=$this->get_html_name($portnum);
    $checked= $value ? ' checked': '';
    return "<input type='checkbox' name='$name' value='1'$checked/>";
  }

/**
 * Overrides to return a false if value is missing.
 * @param int $portnum
 * @return boolean
 */
  function get_value_from_post($portnum=NULL)
  {
    $value=parent::get_value_from_post($portnum);
    return $value?true:false;
  }

}

/* Field_Int *******************************************************************/

class Field_Int extends Field {
  public $min;
  public $max;
  public $unit;

  function __construct($json,$name) {
    parent::__construct($json,$name);
    $this->min = array_key_exists("min",$json)?$json["min"]:NULL;
    $this->max = array_key_exists("max",$json)?$json["max"]:NULL;
    $this->unit = array_key_exists("unit",$json)?$json["unit"]:"";
  }

  function schema_table() {
    parent::schema_table();
    echo "<td>min: $this->min</td><td>max: $this->max</td><td>unit: $this->unit</td>";
  }

  function format_value($value) {
    if ($this->unit) {
      $value.=' '.$this->unit;
    }
    return $value;
  }

  function display_name() {
    $nice = parent::display_name();
    if ($this->min!==NULL && $this->max!==NULL) {
      $nice.= " ($this->min - $this->max)";
    } else if ($this->min!==NULL) {
      $nice.=" (minimum $this->min)";
    } else if ($this->max!==NULL) {
      $nice.=" (maximum $this->max)";
    }
    //if ($this->unit) $nice.= " $this->unit";
    return $nice;
  }

  function edit_html(&$json,$portnum=NULL)
  {
    $value=parent::get_value_from_post($portnum); // neet to get the pre-cast to int version if we're posting back
    if ($value===NULL) {
      $value=$this->read_value($json,$portnum);
    }
    $value=htmlspecialchars($value,ENT_QUOTES);
    $name=$this->get_html_name($portnum);
    $size=6;
    if ($this->max!==NULL) $size=strlen($this->max);
    return "<input type='text' name='$name' value='$value' size='$size'> $this->unit\n";
  }

  function validate($portnum=NULL) {
    $error=NULL;
    $value=parent::get_value_from_post($portnum); // use parent to skip the cast to int
    if (is_numeric($value)) {
      if ($this->max!==NULL && $value>$this->max) $error=lang_d('err_max').' '.$this->max;
      if ($this->min!==NULL && $value<$this->min) $error=lang_d('err_min').' '.$this->min;
    } else {
      $error=lang_d('err_numeric');
    }
    return $error;
  }

/**
 * Override to cast it to an int.
 */
  function get_value_from_post($portnum=NULL) {
    return (int)parent::get_value_from_post($portnum);
  }

}

/* Field_String ****************************************************************/

class Field_String extends Field {
  public $max;

  function __construct($json,$name) {
    parent::__construct($json,$name);
    $this->max = array_key_exists("max",$json)?$json["max"]:"";
  }

  function schema_table() {
    parent::schema_table();
    echo "<td>max: $this->max</td>";
  }

  function edit_html(&$json,$portnum=NULL)
  {
    $value=htmlspecialchars($this->read_value($json,$portnum),ENT_QUOTES);
    $name=$this->get_html_name($portnum);
    $size=20;
    $maxlen='';
    if ($this->max!==NULL) {
      $size=$this->max+1;
      $maxlen=" maxlength='$this->max'";
    }
    return "<input type='text' name='$name' value='$value' size='$size'$maxlen>\n";
  }

  function validate($portnum=NULL) {
    $error=NULL;
    $value=$this->get_value_from_post($portnum);
    if ($this->required && strlen($value)<1) {
      $error=lang_d('err_required');
    }
    return $error;
  }

  function format_value($value)
  {
    return nl2br($value);
  }

}

/* Field_Ipaddress *************************************************************/

class Field_Ipaddress extends Field_String {
  function __construct($json,$name) {
    parent::__construct($json,$name);
  }
  function display_name() {
    $nice = parent::display_name();
    return $nice.' (xxx.xxx.xxx.xxx)';
  }

  function edit_html(&$json,$portnum=NULL)
  {
    $value=htmlspecialchars($this->read_value($json,$portnum),ENT_QUOTES);
    $name=$this->get_html_name($portnum);
    return "<input type='text' name='$name' value='$value' size='16' maxlength='15'>\n";
  }

  function validate($portnum=NULL) {
    $error=NULL;
    $value=$this->get_value_from_post($portnum);
    if (strlen($value)==0 && $this->required) {
      $error=lang_d('err_iprequired');
    }
    if (strlen($value)>0) {
      $nums=explode('.',$value);
      for ($i=0;$i<4;$i++) {
        if (!array_key_exists($i,$nums) || !is_numeric($nums[$i]) || $nums[$i]<0 || $nums[$i]>255) {
          $error= lang_d("err_ipinvalid");
          break;
        }
      }
    }
    return $error;
  }

}

/* Field_Password **************************************************************/

class Field_Password extends Field_String {
  public $nameconfirm;
  public $namecheck;

  function __construct($json,$name) {
    parent::__construct($json,$name);
    $this->nameconfirm=$this->name."_confirm";
    $this->namecheck  =$this->name."_checkbox";
  }

  static function get_html_value_from_post($name) {
    if (array_key_exists($name,$_POST)) {
      return htmlspecialchars($_POST[$name]);
    } else {
      return "";
    }
  }

  static function get_raw_value_from_post($name) {
    if (array_key_exists($name,$_POST)) {
      return $_POST[$name];
    } else {
      return "";
    }
  }

  function edit_html(&$json,$portnum=NULL)
  {
    // If reposting, get display values from post, since the ones we put in the json got encrypted.
    $value=self::get_html_value_from_post($this->name);
    $valueconfirm=self::get_html_value_from_post($this->nameconfirm);
    $valuechecked=self::get_html_value_from_post($this->namecheck);
    $size=20;
    $maxlen='';
    if ($this->max!==NULL) {
      $size=$this->max+1;
      $maxlen=" maxlength='$this->max'";
    }
    $checked=$valuechecked ? ' checked' : '';
    $s="change: <input type='checkbox' name='$this->namecheck' value='changed'$checked><br/>\n";
    if ($this->name=='AdminPassword') {
      $valueold=self::get_html_value_from_post($this->name.'_old');
      $s.="old password: <input type='text' name='{$this->name}_old' value='$valueold' size='$size'$maxlen><br/>\n";
    }
    $s.="new password: <input type='text' name='$this->name' value='$value' size='$size'$maxlen><br/>\n";
    $s.="confirm password: <input type='text' name='$this->nameconfirm' value='$valueconfirm' size='$size'$maxlen><br/>\n";
    return $s;
  }

  function validate($portnum=NULL) {
    $error=NULL;
    $value=self::get_raw_value_from_post($this->name);
    $valueconfirm=self::get_raw_value_from_post($this->nameconfirm);
    if (array_key_exists($this->namecheck, $_POST) && $_POST[$this->namecheck]=='changed') {
      if ($value!=$valueconfirm) {
        $error=lang_d('err_pwmismatch');
      }
    }
    return $error;
  }

  function write_value(&$json,$value,$portnumber=0) {
    // only write the json value if the "changed" checkbox is checked
    if (array_key_exists($this->namecheck, $_POST) && $_POST[$this->namecheck]=='changed') {
      $account = str_replace("Password","",$this->name);
      if ($account == 'Admin')
        {
          $rootpw = $value;
          if ($rootpw != '')
            $rootpw = trim(shell_exec("openssl passwd -1 '$value'"));
          syslog(LOG_INFO,"rootpw='$rootpw'\n");
          $json["network"]["RootPassword"] = $rootpw;
        }
      if ($value != '')
        $value=md5("$account:webapi:$value");
      return parent::write_value($json, $value, $portnumber);
    }
    return false;
  }

  function test_password(&$json,$password) {
    $account = str_replace('Password','',$this->name);
    $value=trim($this->read_value($json));
    $password=trim($password);
    $md5password=md5("$account:webapi:$password");
    return ($value=='') || ($md5password==$value);
  }

}

/* Field_Enum ******************************************************************/

class Field_Enum extends Field {
  public $choices;

  function __construct($json,$name) {
    parent::__construct($json,$name);
    $this->choices = array_key_exists("choices",$json)?$json["choices"]:"";
  }
  function schema_table() {
    parent::schema_table();
    foreach ($this->choices AS $choice) echo "<td>$choice</td>";
  }

  function format_value($value) {
    return lang_d($value);
  }

  function edit_html(&$json,$portnum=NULL) {
    $value=$this->read_value($json,$portnum);
    $name = $this->get_html_name($portnum);
    $html = "<select name='$name'>\n";
    foreach ($this->choices AS $choice) {
      $sel= $choice==$value ? ' selected' : '';
      $choicenice=lang_d($choice);
      $html.=" <option value='$choice'$sel>$choicenice</option>\n";
    }
    $html .="</select>\n";
    return $html;
  }
}

/* Field_Multi *****************************************************************/

class Field_Multi extends Field_Enum {
  function __construct($json,$name) {
    parent::__construct($json,$name);
    if ($this->numlines<count($this->choices)) $this->numlines=count($this->choices);
  }

  function format_value($value) {
    $vals=explode(',',$value);
    $nices=array();
    foreach ($vals as $val) {
      $nices[]=lang_d($val);
    }
    return '<div class="multi-check">'.implode('</div><div class="multi-check">',$nices).'</div>';
  }

  function edit_html(&$json,$portnum=NULL) {
    $value=$this->read_value($json,$portnum);
    $name = $this->get_html_name($portnum);
    $html = "";
    $vals=explode(',',$value);
    foreach ($this->choices AS $choice) {
      $checked= in_array($choice,$vals) ? ' checked' : '';
      $choicenice=lang_d($choice);
      $idName = $name.$choicenice.'ID' ;
      $html[]= "<div class='multi-check'><input type='checkbox' id='$idName' name='{$name}[]' value='$choice'$checked/><label for='$idName'>$choicenice</label></div>";
    }
    return implode(" ",$html);
  }

 /**
 * Overrides to collapse value array into a comma separated string
 * @param int $portnum
 * @return value - NULL if value is not in POST
 */
  function get_value_from_post($portnum=NULL)
  {
    $value=parent::get_value_from_post($portnum);
    if (is_array($value)) {
      $value=implode(',',$value);
    } else {
      $value='';
    }
    return $value;
  }

  function validate($portnum=NULL) {
    $error=NULL;
    $value=$this->get_value_from_post($portnum);
    if (strlen($value)<1 && $this->required) $error=lang_d('err_checkone');
    return $error;
  }
}