<?php

if (!defined('BASEPATH'))
    exit('No direct script access allowed');


/* there's supposed to be a helper function that's loaded to
 * replace the json_indent function from php 5.2, but that isn't
 * getting loaded (don't know why.  So we'll put in a stub for now
 */
if (!function_exists('json_indent')) {
  function json_indent($json) {
    return $json;
  }
}

/**
 * This holds the data for a "page" of the UI.
 */
class Page {

    public $pagename;    // string- name of page
    public $source;      // string- "file" if getting json directly from a file.
    public $filename;    // string- name of file to load.
    public $readcommand; // CLI command to read data file into stdout
    public $writecommand; // CLI command to write data file from stdin
    public $verifycommand; // CLI command to validate page data prior to saving.
    public $resetcommand; // CLI command to reset page
    public $updaterate;  // seconds between automatic refresh. If 0 or absent, don't refresh.
    public $json;        // json array of values
    public $portfields;  // array[fieldname] of Field
    public $fields;      // array[fieldname] of Field
    public $verticalfields; // array[fieldname] of Field
    public $verticaladdress; // address of vertical array in JSON
    public $porterrors;  // array[fieldname][portnum] of string
    public $fielderrors; // array[fieldname] of string
    public $pageerrors;  // strings returned from write commands or other page-wide errors.
    public $fieldsupdated; // array of passwords set in field_password

    /**
     * Loads the schema for this page from JSON. Passes fragments to fields as it creates them.
     * @param array $json - portion of the JSON schema file for this page.
     */

    public function __construct($json, $pagename) {
        $this->pagename = $pagename;
        $this->source = array_key_exists("source", $json) ? $json['source'] : '';
        $this->filename = array_key_exists("filename", $json) ? $json['filename'] : '';
        $this->readcommand = array_key_exists("readcommand", $json) ? $json['readcommand'] : '';
        $this->writecommand = array_key_exists("writecommand", $json) ? $json['writecommand'] : '';
        $this->verifycommand = array_key_exists("verifycommand", $json) ? $json['verifycommand'] : '';
        if ($this->verifycommand == '' && is_array($this->writecommand)) $this->verifycommand=array();
        $this->resetcommand = array_key_exists("resetcommand", $json) ? $json['resetcommand'] : '';
        $this->updaterate = array_key_exists("updaterate", $json) ? $json['updaterate'] : 0;
        $this->verticaladdress = array_key_exists("verticaladdress", $json) ? $json['verticaladdress'] : 'vertical';
        $this->portfields = Array();

        // HBB - This to load up the PHP field objects for each schema portfield entries
        //     - These PHP field objects performs the operations on the field depending on the type
        //       from the schema field descriptor.
        if (array_key_exists("portfields", $json)) {
            foreach ($json["portfields"] AS $fieldname => $fieldjson) {
            		$this->portfields[$fieldname] = Field::createFromJson($fieldname, $fieldjson);
            }
        }
        $this->fields = Array();
        if (array_key_exists("fields", $json)) {
            foreach ($json["fields"] AS $fieldname => $fieldjson) {
            	$this->fields[$fieldname] = Field::createFromJson($fieldname, $fieldjson);
            }
        }
        $this->verticalfields = Array();
        if (array_key_exists("verticalfields", $json)) {
            foreach ($json["verticalfields"] AS $fieldname => $fieldjson) {
                $this->verticalfields[$fieldname] = Field::createFromJson($fieldname, $fieldjson);
            }
        }
        $this->porterrors = array();
        $this->fielderrors = array();
        $this->pageerrors = '';
        $this->fieldsupdated=array();
    }

    /**
     * Dump of the schema for testing
     */
    public function schema_table() {
        foreach ($this->fields AS $field) {
            echo "<tr>";
            $field->schema_table();
            echo "</tr>";
        }
        foreach ($this->portfields AS $portfield) {
            echo "<tr>";
            $portfield->schema_table();
            echo "</tr>";
        }
    }

    /**
     * Returns a reference to the location specified by $address in the JSON array.
     * @param string $address
     * @return reference to an array or scalar in the JSON
     */
    public function get_json_value($address) {
        $a = explode('.', $address);
        $j = &$this->json;
        foreach ($a AS $index) {
            if ($index == '[*]')
                $index = $portnumber;
            if ($index == '[?]')
                $index = $verticalnumber;
            if (array_key_exists($index, $j)) {
                $j = &$j[$index];
            } else {
                return '';
            }
        }
        return $j;
    }

    public function get_numPorts(){
      return Pages_model::$num_ports;
    }

  public function get_value_field_rows()
  {
    $s="";
    foreach ($this->fields AS $field) {
      $nicename=$field->display_name();
      $value=$field->format_value($field->read_value($this->json));
      $s.= "<tr><td><b>$nicename</b></td><td>$value</td></tr>\n";
    }
    return $s;
  }

  public function get_value_port_rows()
  {
    $s="";
    $num_ports=Pages_model::$num_ports;
    foreach ($this->portfields AS $fieldname=>$field) {
      $nicename=$field->display_name();
      $s.= "<tr><td><b>$nicename</b></td>";
      for ($portnum=0;$portnum<$num_ports;$portnum++) {
        $value=$field->format_value($field->read_value($this->json,$portnum));
        $s.= "<td>$value</td>";
      }
      $s.= "</tr>\n";
    }
    return $s;
  }

  public function get_value_vertical_rows()
  {
    $num_verticals=count($this->get_json_value($this->verticaladdress));
    $s= "<tr>";
    foreach ($this->verticalfields AS $fieldname=>$field) {
      $nicename=$field->display_name();
      $s.= "<th>$nicename</th>";
    }
    echo "</tr>";
    for ($verticalnum=0;$verticalnum<$num_verticals;$verticalnum++) {
      $s.= "<tr>";
      foreach ($this->verticalfields AS $fieldname=>$field) {
        $value=$field->format_value($field->read_value($this->json,0,$verticalnum));
        $s.= "<td>$value</td>";
      }
      $s.= "</tr>\n";
    }
    return $s;
  }

    public function get_edit_field_rows()
  {
    $s="";
    foreach ($this->fields AS $fieldname=>$field) {
      $nicename=$field->display_name();
      $html=$field->edit_html($this->json);
      $err='';
      if (array_key_exists($field->name,$this->fielderrors)) {
        $errmessage=$this->fielderrors[$field->name];
        $err="<br/><span style='color:red'>$errmessage</span>";
      }
      $valid = $err ? '<span  style="color:red">' : '<span>';
      $s.= "<tr><td>{$valid}{$nicename}</span></td><td>$html$err</td></tr>\n";
    }
    return $s;
  }

  public function get_edit_vertical_rows()
  {
    // not used- sessions and connections aren't editable
  }


    /**
     * Network Page display for values
     */
    public function view_value_table() {
        // TODO: Move this to the view
        //echo "<table><thead><tr><th>Network Configuration Page parameters</th><th>Values</th></tr></thead><tbody>\n";
        return $this->get_value_field_rows();
        //echo "</tbody></table>\n";
    }

     /**
     * Network Page display for values
     */
    public function view_edit_table() {
        // TODO: Move this to the view, and wrap it in "form" tags
        //echo "<table><thead><tr><th>Network Configuration Page parameters</th><th>Values</th></tr></thead><tbody>\n";
        return $this->get_edit_field_rows();
        //echo "</table>\n";
    }


    public function get_port_row_headers() {
        $strReturn = '';
        $altRow = false;
        $i = 0;
        foreach ($this->portfields as $portfield) {
			// HBB - added 12/27/2012
        	if (($portfield->entryList == "cont") ||
        			($portfield->entryList == "end"))
        	{
				// HBB - no sidebar name for continue and end of entry list
        		continue;
        	}
        	$alt = $altRow ? 'alt-port-header' : '';
            $altRow = !$altRow;
            $hidemaster = $portfield->hidemaster ? 'hide-master' : '';
            $numlines = " num-lines-{$portfield->numlines}" ;
            $hidegroup = $portfield->hidegroup ? "hidden hide-$portfield->hidegroup " : '';
            if  ( isset($portfield->level) ){
                $hidegroup  = 'sub-header-'.$portfield->level.' ' ;
            }
             //$multi = $portfield->type == 'multi' ? " multi-" . count($portfield->choices) : '';
            $strReturn .= "<li class='$hidegroup port-$portfield->type sub-info $numlines $alt $hidemaster'>";
            if ($portfield->hidemaster) {
                $i++;
                $strReturn .= "<a href='#' class='sb-show hidden-toggle' id='section-".$i."' onclick=\"hide_show('hide-$portfield->hidemaster')\">+</a>";
            }
            $strReturn .= $portfield->display_name();
            $strReturn .= "</li>\n";
        }
        return $strReturn;
    }

    public function get_port_column($portNum, $maxports=4) {
        $strReturn = '';
        // HBB - added 11/13/12 - make sure the portNum is valid
        if ((0 > $port) || ($port >= $maxports))
        {
            syslog(LOG_INFO, "ERROR: get_port_column() invalid portNum $portNum\n");
        	return $strReturn;
        }

        $altRow = false;
        foreach ($this->portfields as $portfield) {
            // HBB - rewrote to add entry sequence list on 12/27/2012
            if ($portfield->entryList == "start") {
            	$alt = $altRow ? 'alt-port-column' : '';
            	$altRow = !$altRow;
            	// this starts the entry list
            	$hidegroup = $portfield->hidegroup ? "hidden hide-$portfield->hidegroup " : '';
            	if  ( isset($portfield->level) ){
            		$hidegroup = 'sub-header-'.$portfield->level.' ' ;
            	}
            	$numlines = "num-lines-{$portfield->numlines}" ;
             	$strReturn .= "<li class='$alt $numlines {$hidegroup}port-$portfield->type port-view sub-info' >";
            	$strReturn .= $portfield->format_value($portfield->read_value($this->json, $portNum));
           		$strReturn .= "</class>";
             } else if ($portfield->entryList == "cont") {
            	// this continues the entry list
            	$hidegroup = $portfield->hidegroup ? "hidden hide-$portfield->hidegroup " : '';
            	if  ( isset($portfield->level) ){
            		$hidegroup = 'sub-header-'.$portfield->level.' ' ;
            	}
            	$numlines = "num-lines-{$portfield->numlines}" ;
            	$strReturn .= "<class='$alt $numlines {$hidegroup}port-$portfield->type port-view sub-info' >";
            	$strReturn .= " ";
            	$strReturn .= $portfield->format_value($portfield->read_value($this->json, $portNum));
            	$strReturn .= "</class>";
            } else if ($portfield->entryList == "end") {
            	// this ends the entry list and goes to next row
            	$hidegroup = $portfield->hidegroup ? "hidden hide-$portfield->hidegroup " : '';
            	if  ( isset($portfield->level) ){
            		$hidegroup = 'sub-header-'.$portfield->level.' ' ;
            	}
            	$numlines = "num-lines-{$portfield->numlines}" ;
            	$strReturn .= "<class='$alt $numlines {$hidegroup}port-$portfield->type port-view sub-info' >";
            	$strReturn .= " ";
            	$strReturn .= $portfield->format_value($portfield->read_value($this->json, $portNum));
            	$strReturn .= "</class>";
            	$strReturn .= "</li>\n";
            } else {
            	// standard way of handling an entry when there is no list - one entry field per
            	$alt = $altRow ? 'alt-port-column' : '';
            	$altRow = !$altRow;
            	$hidegroup = $portfield->hidegroup ? "hidden hide-$portfield->hidegroup " : '';
            	if  ( isset($portfield->level) ){
            		$hidegroup = 'sub-header-'.$portfield->level.' ' ;
            	}
            	$numlines = "num-lines-{$portfield->numlines}" ;
            	$strReturn .= "<li class='{$hidegroup}port-$portfield->type port-view sub-info $numlines $alt' >";
            	$strReturn .= $portfield->format_value($portfield->read_value($this->json, $portNum));
            	$strReturn .= "</class></li>\n";
            }
        }
        return $strReturn;
    }

    public function get_port_column_edit($portNum) {
        $strReturn = '';
        $altRow = false;

        foreach ($this->portfields as $portfield) {
            $liClasses = array();
            $err = "";
            if (array_key_exists($portfield->name,$this->porterrors) && array_key_exists($portNum,$this->porterrors[$portfield->name])) {
                $errmessage=$this->porterrors[$portfield->name][$portNum];
                $err.="<span class='option-err-msg'>$errmessage</span>";
            }
            $liClasses[] = $altRow ? 'alt-port-edit' : '';
            if (($portfield->entryList == "start") ||
            		($portfield->entryList == "none"))
            {
            	$altRow = !$altRow;
            }
            $liClasses[] = $portfield->hidegroup ? "hidden hide-$portfield->hidegroup" : '';
            if  ( isset($portfield->level) ){
                $liClasses[] = 'sub-header-'.$portfield->level ;
            }
            $liClasses[] = "num-lines-{$portfield->numlines}" ;
            $liClasses[] = "port-$portfield->type port-edit sub-info";
            $classStr = '';
            foreach ($liClasses as $liClass){
                if ($liClass){ $classStr .= $liClass.' '; }
            }
            // HBB - added entry list capability 12/27/2012
            if ($portfield->entryList == "start") {
            	// this starts the entry list
            	$strReturn .= "<li class='$classStr' >";
            	$strReturn .= $portfield->edit_html($this->json, $portNum);
            	$strReturn .= "</class='$classStr' >";
            } else if ($portfield->entryList == "cont") {
            	// this continues the entry list
            	$strReturn .= "<class='$classStr' >";
            	$strReturn .= " ";
            	$strReturn .= $portfield->edit_html($this->json, $portNum);
            	$strReturn .= "</class='$classStr' >";
            } else if ($portfield->entryList == "end") {
            	// this ends the entry list and goes to next row
            	$strReturn .= "<class='$classStr' >";
            	$strReturn .= " ";
            	$strReturn .= $portfield->edit_html($this->json, $portNum);
            	$strReturn .= "</class='$classStr' >";
            	$strReturn .= $err."</li>\n";
            } else {
            	// standard way of handling an entry when there is no list - one entry field per
            	$strReturn .= "<li class='$classStr' >";
            	$strReturn .= $portfield->edit_html($this->json, $portNum);
            	$strReturn .= $err."</li>\n";
            }
        }
        if ( isset($this->pageerrors) && $this->pageerrors  ){
            $strReturn .= "<li class='port-err-msg'>" . nl2br($this->pageerrors) . "</li>";
        }
        return $strReturn;
    }

     public function get_edit_port_rows()
  {
    $s="";
    $num_ports = Pages_model::$num_ports;
    foreach ($this->portfields AS $fieldname=>$field) {
      $nicename=$field->display_name();
      $s.= "<tr><td>$nicename</td>";
      for ($portnum=0;$portnum<$num_ports;$portnum++) {
        $html=$field->edit_html($this->json,$portnum);
        $err='';
        if (array_key_exists($field->name,$this->porterrors) && array_key_exists($portnum,$this->porterrors[$field->name])) {
          $errmessage=$this->porterrors[$field->name][$portnum];
          $err="<br/><span style='color:red'>$errmessage</span>";
        }
        $s.= "<td>$html$err</td>";
      }
      $s.= "</tr>\n";
    }
    return $s;
  }

    /**
     * Updates all fields from POST that are present into the $this->json array so it can be saved.
     * @param int portnum - optional port number. If not-null, only loads field from that port.
     */
    public function update_from_post($portnum=NULL) {
        if ($portnum === NULL) {
            $num_ports = Pages_model::$num_ports;
            foreach ($this->fields AS $field) {
                $value = $field->get_value_from_post();
                if ($value !== NULL) {
                    if ($field->write_value($this->json, $value)) {
                        $this->fieldsupdated[]=$field->name;
                    }
                }
            }
            foreach ($this->portfields AS $portfield) {
                for ($portnum = 0; $portnum < $num_ports; $portnum++) {
                    $value = $portfield->get_value_from_post($portnum);
                    if ($value !== NULL) {
                        $portfield->write_value($this->json, $value, $portnum);
                    }
                }
            }
        } else { // only save one port.
            foreach ($this->portfields AS $portfield) {
                $value = $portfield->get_value_from_post($portnum);
                if ($value !== NULL) {
                    $portfield->write_value($this->json, $value, $portnum);
                }
            }
        }
    }

    /**
     * Executes a shell command that returns a JSON file in stdout, and returns the result as a json array.
     * @param string $command
     * @return array
     */
    private static function load_command($command) {
        $file = shell_exec($command);
        $json = json_decode($file, true);
        if (!$json) {
            $json = array();
        }
        return $json;
    }

    /**
     * Encodes a json array into a string, and then feeds it to a shell command that takes a file from stdin.
     * If the command returns non-zero, appends stderr to the $errors string.
     * returns the return code if there's an error.
     * @param string $command
     * @param array $json
     * @param string $errors
     * @return int - 0 if successful
     */
    private static function save_command($command, &$json, &$errors) {
        if (trim($command)=='') return 0; // if it's null, it's successfull.
        $descriptorspec = array(
            0 => array("pipe", "r"),
            1 => array("pipe", "w"),
            2 => array("pipe", "w")
        );
        $process = proc_open($command, $descriptorspec, $pipes);
        if (is_resource($process)) {
            $json_file = json_indent(json_encode($json));
            fwrite($pipes[0], $json_file);
            fclose($pipes[0]);

            $outputstr = stream_get_contents($pipes[1]);
            fclose($pipes[1]);
            $errstr = stream_get_contents($pipes[2]);
            fclose($pipes[2]);
            $return = proc_close($process);
            if ($return) {
                if ($errstr) {
                    $errors.=$errstr;
                } else {
                    $errors.="'$command' returned error code '$return'.\n";
                }
            }
            return $return;
        }
        return 1;
    }

    /**
     * Loads $this->json array from what's in the page schema.
     * That means loading up the JSON file by executing the command(s) in "readcommand".
     */
    public function load() {
        $num_ports = Pages_model::$num_ports;
        if ($this->source == 'file') {
            $json_file = file_get_contents($this->filename);
            $this->json = json_decode($json_file, true);
        } else if ($this->readcommand) {
            if (is_array($this->readcommand)) {
                $this->json = Array();
                foreach ($this->readcommand AS $cmd) {
                    $this->json[] = self::load_command($cmd);
                }
            } else {
                $this->json = self::load_command($this->readcommand);
            }
        } else {
            throw new Exception("page source other than file not implemented");
        }
        if (!$this->json) {//throw new Exception("Error reading $this->filename");
            $this->json = array();
        }
    }

    /**
     * Encodes $this->json array back into JSON and writes the file(s) where it belongs
     * by executing the command(s) in "writecommand".
     */
    public function save() {
        $innocent = true;
        $this->pageerrors = '';
        $num_ports = Pages_model::$num_ports;
        $json_file = json_indent(json_encode($this->json));
        if ($this->source == 'file') {
            if (!file_put_contents($this->filename, $json_file)) {
                $guilty = true;
                $this->pageerrors = "Error writing file $this->filename\n";
            }
        } else if ($this->writecommand) {
            if (is_array($this->writecommand)) {
                for ($i = 0; $i < count($this->writecommand); $i++) {
                    if (array_key_exists($i,$this->verifycommand) && self::save_command($this->verifycommand[$i], $this->json[$i], $this->pageerrors)) {
                         $innocent = false;
                    } else if (self::save_command($this->writecommand[$i], $this->json[$i], $this->pageerrors)) {
                        $innocent = false;
                    }
                }
            } else {
                if (self::save_command($this->verifycommand, $this->json, $this->pageerrors)) {
                    $innocent = false;
                } else if (self::save_command($this->writecommand, $this->json, $this->pageerrors)) {
                    $innocent = false;
                }
            }
        } else {
            throw new Exception("page source other than file not implemented");
        }
        return $innocent;
    }

    /**
     * Checks all fields in the page for validity.
     * @param int $portnum - optional port number. If not-null, only checks that port.
     * @return boolean- true if no errors.
     * side effect- sets the $fielderrors and $porterrors arrays with error strings
     */
    public function validate($portnum=NULL) {
        $innocent = 1;
        if ($portnum === NULL) {
            $num_ports = Pages_model::$num_ports;
            foreach ($this->fields AS $field) {
                $error = $field->validate();
                if ($error !== NULL) {
                    $this->fielderrors[$field->name] = $error;
                    $innocent = 0;
                }
            }
            foreach ($this->portfields AS $portfield) {
                for ($portnum = 0; $portnum < $num_ports; $portnum++) {
                    $error = $portfield->validate($portnum);
                    if ($error !== NULL) {
                        $this->porterrors[$portfield->name][$portnum] = $error;
                        $innocent = 0;
                    }
                }
            }
        } else {
            foreach ($this->portfields AS $portfield) {
                $error = $portfield->validate($portnum);
                if ($error !== NULL) {
                    $this->porterrors[$portfield->name][$portnum] = $error;
                    $innocent = 0;
                }
            }
        }
        return $innocent;
    }

    public function reset()
    {
      if (!$this->resetcommand) return;
      if (is_array($this->resetcommand)) {
        foreach ($this->resetcommand AS $command) {
          if ($command) {
            exec($command);
          }
        }
      } else {
        exec($this->resetcommand);
      }
    }

}
