<?php if ( ! defined('BASEPATH')) exit('No direct script access allowed');


class Software_manager {
  public $config;
  public $cmdpath;
  public $errors;

  function __construct() {
    $this->config=json_decode(file_get_contents("assets/json/software_schema.json"),true);
    $this->cmdpath = array_key_exists('cmdpath', $this->config) ? $this->config['cmdpath'] : '';
    $this->errors = '';
  }

/**
 * Gets and clears the accumulated error string from previous commands.
 * @return string
 */
  public function get_errors() {
    $errs=$this->errors;
    $this->errors='';
    return $errs;
  }

/**
 * Parses various version strings to get the version and partition name.
 * Can get "name - version -", which is ipkg format, "name version", or just "version".
 * @param string $outputstr - string to parse.
 * @param string $partname - reference to partition name variable
 * @return string - version string
 */
  static function parse_version($outputstr,&$partname) {
    $firstdash=strpos($outputstr,' - ');
    if ($firstdash>0) {
      $seconddash=strpos($outputstr,' -',$firstdash+1);
      if ($seconddash) {
        $partname=substr($outputstr,0,$firstdash);
        $version=substr($outputstr,$firstdash+3,$seconddash-$firstdash-3);
      } else {
        // just get the rest of the line
        $version=substr($outputstr,$firstdash+3);
      }
    } else {
      $lastspace=strrpos($outputstr,' ');
      if ($lastspace===false) {
        $version=$outputstr;
        $partname="";
      } else {
        $version=substr($outputstr,$lastspace+1);
        $partname=substr($outputstr,$lastspace);
      }
    }
    return $version;
  }

/**
 * Utility function to call a shell command with STDIN, STDOUT, and STDERR
 * If any errors appends SDTERR to the accumulated error string $this->errors
 * @param string $command
 * @param string $inputstr  - string to send to STDIN
 * @param string $outputstr   - string from STDOUT
 * @return int- return value from shell command
 */
  private function exec_files($command,$inputstr,&$outputstr)
  {
    $descriptorspec = array(
        0 => array("pipe", "r"),
        1 => array("pipe", "w"),
        2 => array("pipe", "w")
    );
    $process = proc_open($this->cmdpath.$command, $descriptorspec, $pipes);
    if (is_resource($process)) {
      if ($inputstr) {
        fwrite($pipes[0], $inputstr);
      }
      fclose($pipes[0]);

      $outputstr = stream_get_contents($pipes[1]);
      fclose($pipes[1]);
      $errstr = stream_get_contents($pipes[2]);
      fclose($pipes[2]);
      $return = proc_close($process);
      if ($return) {
        if ($errstr) {
          $this->errors .= $errstr."\n";
        } else {
          $this->errors .= "'$command' returned error code '$return'.\n";
        }
      }
      return $return;
    }
    return 1;
  }

/**
 * Reboots the device. Returns true for success.
 * @return boolean
 */
  function reboot()
  {
    $command=$this->config['commands']['reboot'];
    if ($this->exec_files($command, '', $outputstr)) {
      $innocent=false;
    } else {
      $innocent=true;
    }
    return $innocent;
  }

/**
 * Re-initializes a port to default values. Returns true for success.
 * @param int portnum
 * @return boolean
 */
  function reinitialize_port($portnum)
  {
      $command=$this->config['commands']['reinitialize_port'];
      $command=str_replace('<port>',$portnum,$command);
      if ($this->exec_files($command, '', $outputstr)) {
        $innocent=false;
      } else {
        $innocent=true;
      }
      return $innocent;
  }

/**
 * returns an array of partition version strings, indexed by partition name
 * @return array or false if errors
 */
  function getPartitionVersions()
  {
    $versions=array();
    foreach ($this->config['partitions'] AS $partname=>$part) {
      $outputstr='';
      $command=$part['partition_version_cmd'];
      $command=str_replace('<partition>',$part['partition'],$command);
      if ($this->exec_files($command, '', $outputstr)) {
        $versions[$partname] = '<unknown>';
      } else {
        $partit='';
        $version = self::parse_version($outputstr,$partit);
        $versions[$partname] = trim($version);
      }
    }
    return $versions;
  }

/**
 * returns array of application version strings, indexed by app name
 * @return array
 */
  function getApplicationVersions()
  {
    $output='';
    $this->exec_files($this->config['commands']['application_versions'],'',$output);
    $arr=explode("\n",$output);
    $appversions=array();
    foreach ($arr AS $line) {
      if (trim($line)) { // ignore blank lines
        $appname='';
        $version=self::parse_version($line,$appname);
        $appversions[$appname]=$version;
      }
    }
    return $appversions;
  }

/**
 * Verifies an image file to be installed into a partition
 * @param string $file  - name of file to install
 * @param string $partitionname - name of partition to install, from the schema
 * @return boolean- true if success
 */
  function verifyImageFile($file,$partitionname)
  {
    if (!array_key_exists($partitionname, $this->config['partitions'])) {
      $this->errors.="Invalid partition '$partitionname'";
      return false;
    }
    $output='';
    $this->errors='';
    $command=$this->config['partitions'][$partitionname]['verify_image_cmd'];
    $command=str_replace('<file>', $file, $command);
    $result=$this->exec_files($command,'',$output);
    if ($result && trim($this->errors)=='') {
      $this->errors = "Error in verify command '$command'.\n";
    }
    return $result<1;
  }

/**
 * Installs a file into a partition
 * @param string $file  - name of file to install
 * @param string $partitionname - name of partition to install, from the schema
 * @return boolean- true if success
 */
  function installImageFile($file,$partitionname)
  {
    $output='';
    $this->errors='';
    $command=$this->config['partitions'][$partitionname]['write_image_cmd'];
    $command=str_replace('<file>', $file, $command);
    $partition=$this->config['partitions'][$partitionname]['partition'];
    $command=str_replace('<partition>', $partition, $command);
    $result=$this->exec_files($command,'',$output);
    if ($result && trim($this->errors)=='') {
      $this->errors = "Error in image install command '$command'.\n";
    }
    return $result<1;
  }

/**
 * Checks that the version string is more than 2 numbers and that the first two are the same as the app base.
 * @param string $versionstring - string returned from ipkCheck
 * @return boolean  - true of it's valid
 */
  function checkApplicationVersion($versionstring) {
    $this->errors = "";
    $versions=$this->getApplicationVersions();
    $mgrversion=$versions['application-base'];
    $mgrarr=explode('.',$mgrversion);
    $partname='';
    $appversion=self::parse_version($versionstring,$partname);
    $apparr=explode('.',$appversion);
    if (count($apparr)<=2) {
      $this->errors = "Invalid version string in application file.\n";
      return false;
    }
    if ($apparr[0]!=$mgrarr[0] || $apparr[1]!=$mgrarr[1]) {
      $this->errors = "Incompatible application version.\n";
      return false;
    }
    return true;
  }

/**
 * Verifies an ipkg file to be installed
 * @param string $file - name of file to install
 * @return boolean - true if it's a valid file
 */
  function verifyApplication($file)
  {
    $output='';
    $this->errors = '';
    $command=$this->config['commands']['application_verify'];
    $command=str_replace('<file>', $file, $command);
    $result=$this->exec_files($command,'',$output);
    if ($result && trim($this->errors)=='') {
      $this->errors = "Error in verify command '$command'.\n";
    }
    if (!$result) {
      // check application version
      return $this->checkApplicationVersion($output);
    }
    return $result<1;
  }

/**
 * Installs an ipkg file
 * @param string $file - name of file to install
 * @return boolean - true if it installed without error
 */
  function installApplication($file)
  {
    $output='';
    $this->errors = '';
    $newfile = $file . '.ipk';
    rename($file, $newfile);
    $command=$this->config['commands']['application_install'];
    $command=str_replace('<file>', $newfile, $command);
    $result=$this->exec_files($command,'',$output);
    if ($result && trim($this->errors)=='') {
      $this->errors = "Error in application install command '$command'.\n";
    }
    return $result<1;
  }

/**
 * Installs an ipkg file in the background after delaying 1 second
 * @param string $file - name of file to install
 * @return boolean - true if it installed without error
 */
  function installApplicationDelayed($file)
  {
    $output='';
    $this->errors = '';
    $newfile = $file . '.ipk';
    rename($file, $newfile);
    $command=$this->config['commands']['application_install_delayed'];
    $command=str_replace('<file>', $newfile, $command);
    $result=$this->exec_files($command,'',$output);
    if ($result && trim($this->errors)=='') {
      $this->errors = "Error in application install command '$command'.\n";
    }
    return $result<1;
  }

/**
 * Lists log files
 * @return array(string)
 */
  function listLogFiles()
  {
    return $this->config['logfiles'];
  }

/**
 * Gets a log file, or false if error.
 * @param string $name
 * @return string
 */
  function getLogFile($name)
  {
    $output='';
    $command=$this->config['commands']['get_log_file'];
    $command=str_replace('<name>',$name,$command);
    $result=$this->exec_files($command,'',$output);
    if ($result && trim($this->errors)=='') {
      $this->errors.="Error in get log command.\n";
    }
    return $result==0 ? $output : false;
  }

/**
 * Clears a log file, or false if error.
 * @param string $name
 * @return boolean
 */
  function clearLogFile($name)
  {
    $output='';
    $command=$this->config['commands']['clear_log_file'];
    $command=str_replace('<name>',$name,$command);
    $result=$this->exec_files($command,'',$output);
    if ($result && trim($this->errors)=='') {
      $this->errors.="Error in clear log command.\n";
    }
    return $result==0;
  }
}

?>